from scm.plams import *
import multiprocessing, os, sys
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from natsort import natsorted


"""
Script that (mostly) automates the method described in https://www.scm.com/doc.trunk/Tutorials/MolecularDynamicsAndMonteCarlo/ViscosityBenzene.html
for calculating the viscosity of Benzene (or any other system). The script implements a ViscosityCalculator object that contains all of the necessary 
methods for calculating the viscosity. All analysis methods simply require the KFFile so they can be used for previously run MD simulations.
"""


class ViscosityCalculator:
    def __init__(self, steps=100_000, sampling_freq=5, time_step=1, logfile=sys.stdout):
        # We define settings here so that after initializing the object the settings can easily be accessed and changed.
        self.MDdefaults = Settings()
        self.MDdefaults.input.ams.Task = "MolecularDynamics"
        self.MDdefaults.input.ams.MolecularDynamics.NSteps = steps
        self.MDdefaults.input.ams.MolecularDynamics.TimeStep = time_step  # in fs
        self.MDdefaults.input.ams.MolecularDynamics.Trajectory.SamplingFreq = sampling_freq  # in steps
        self.MDdefaults.input.ams.MolecularDynamics.Trajectory.WriteVelocities = "No"
        self.MDdefaults.input.ams.MolecularDynamics.Trajectory.WriteCharges = "No"
        self.MDdefaults.input.ams.MolecularDynamics.Trajectory.WriteBonds = "No"
        self.MDdefaults.input.ams.MolecularDynamics.Trajectory.WriteMolecules = "No"
        self.MDdefaults.input.ams.MolecularDynamics.Thermostat.Type = "NHC"
        self.MDdefaults.input.ams.MolecularDynamics.Thermostat.Tau = 100
        self.MDdefaults.input.ams.MolecularDynamics.CalcPressure = "Yes"
        self.MDdefaults.input.ams.MolecularDynamics.Checkpoint.Frequency = 100000
        self.MDdefaults.input.ForceField.Type = "GAFF"
        self.MDdefaults.input.ForceField.AntechamberIntegration = "Yes"

        self.Analysisdefaults = Settings()
        self.Analysisdefaults.input.Task = "AutoCorrelation"
        self.Analysisdefaults.input.AutoCorrelation.Property = "PressureTensor"
        self.Analysisdefaults.input.AutoCorrelation.VecElements.Index = [4, 5, 6]
        self.Analysisdefaults.input.AutoCorrelation.MaxFrame = int(
            20000 / (time_step * sampling_freq)
        )  # each step around 15 ps
        self.Analysisdefaults.input.TrajectoryInfo.Trajectory.Range = f"{int(steps*0.1)} {steps+1}"

    def __call__(self, *args, **kwargs):
        return self.get_viscosities(*args, **kwargs)

    def get_md_KFFiles(self, directory):
        # Method that searches for ams.rkf from md runs in specified directory
        files = [str(p.resolve()) for p in Path(directory).rglob("ams.rkf")]
        return natsorted(files)

    def _get_md_job(self, mol, temperature, name):
        s = self.MDdefaults.copy()
        s.input.ams.MolecularDynamics.InitialVelocities.Temperature = temperature
        s.input.ams.MolecularDynamics.Thermostat.Temperature = temperature
        return AMSJob(name=name, molecule=mol, settings=s)

    def _get_analysis_job(self, KFFilename, name):
        s = self.Analysisdefaults.copy()
        s.input.TrajectoryInfo.Trajectory.KFFilename = KFFilename
        return AMSAnalysisJob(name=name, settings=s)

    def analyse_md(self, MDKFFilepaths, job_names=None, do_init=True, name="viscosity_analysis", calculation_path=None):
        """
        Method that implements the analysis for multiple MD runs
        steps:
        - obtain temperatures
        - get lattice vectors and calculate volume
        - run ACF calculations
        - analyse ACF integrals
        - calculate the viscosities

        arguments:
        MDKFFilepaths      - paths of the KFFiles for the MD runs
        job_names        - name for the analysis run, takes 'title' property from the KFFiles if None
        do_init          - whether to call plams.init at the start. Should be True unless already called
        name             - name for plams.init only necessary when do_init=True
        calculation_path - PLAMS workdirectory for when do_init=True
        """

        if do_init:
            init(path=calculation_path, folder=name)

        workdir = config.default_jobmanager.workdir
        MDtemperatures = [self.get_MD_temperature(f) for f in MDKFFilepaths]

        readers = [KFReader(f) for f in MDKFFilepaths]
        if job_names is None:
            job_names = [r.read("General", "title") + "_analysis" for r in readers]

        lattice_vectors = [
            np.asarray([float(x) for x in r.read("InputMolecule", "LatticeVectors")]).reshape(3, 3) for r in readers
        ]
        bohr2ang = Units.conversion_ratio("bohr", "angstrom")
        volumes = [abs(np.linalg.det(l)) * bohr2ang**3 for l in lattice_vectors]

        ##### ANALYSIS RUNS
        print("Starting analysis runs ...")
        analysisresults = [self._get_analysis_job(f, name).run() for f, name in zip(MDKFFilepaths, job_names)]
        # wait for runs to finish. AMSAnalysisResults does not have method ok so we implement it ourself
        while not all(r.job.check() for r in analysisresults):
            pass

        # once done grab the rkf files
        analysisrkf = [r[f"{r.job.name}.kf"] for r in analysisresults]

        # read the ACF integral and guess the viscosities
        kB = Units.constants["k_B"]
        ACF = [self.get_ACF_integral(rkf) for rkf in analysisrkf]  # in Pa^2 s from au^2 fs
        viscosities = [
            V * 1e-30 / (kB * Tobs[0]) * integral[0] for V, Tobs, integral in zip(volumes, MDtemperatures, ACF)
        ]

        # also plot the viscosities vs time
        self.plot_viscosities(volumes, MDtemperatures, analysisrkf, out_path=os.path.join(workdir, "viscosities.png"))

        if do_init:
            finish()

        temps = [t[0] for t in MDtemperatures]
        temps_sd = [t[1] for t in MDtemperatures]

        # sort everything according to decreasing temps
        si = np.argsort(temps)
        temps = [temps[i] for i in si]
        temps_sd = [temps_sd[i] for i in si]
        viscosities = [viscosities[i] for i in si]

        results = {
            "temperatures": temps,  # in K
            "temperature_stdevs": temps_sd,  # in K
            "viscosities": viscosities,  # in Pa s
            "volumes": volumes,
        }

        return results

    def get_MD_temperature(self, MDKFFilepath, steps_fraction=0.4):
        """
        Method used for extracting the simulation temperature during MD runs.
        Simply takes the average temperatures over a number of steps

        arguments:
        MDKFFilepath - path to the KFFile of the MD run
        steps_fraction - number of steps to use for the averaging. Default 0.4 uses 40% of last steps
        """

        reader = KFReader(MDKFFilepath)
        MDtemperatures = np.fromiter(KFHistory(reader, "MDHistory").iter("Temperature"), float)
        average = np.mean(MDtemperatures[int(-MDtemperatures.size * steps_fraction) :])
        sd = np.std(MDtemperatures[int(-MDtemperatures.size * steps_fraction) :])
        return average, sd

    def get_flat_areas(self, y, low_step=0, high_step=-1, kernel_size=500, eps_size=0.33, split_tolerance=10):
        # prune input y into provided region
        y = y[low_step:high_step]
        # first obtain the difference vector, which is analogous to the first derivative and will allow us to detect flat areas around 0
        diff_y = np.diff(y)
        # we have a lot of noise in our data, so we will calculate the rolling average with a certain kernel_size
        filt = np.ones(kernel_size) / kernel_size
        conv_diff_y = np.convolve(diff_y, filt, mode="valid")
        # We now want to detect areas that are close to zero in the difference vector
        # we need an epsilon to compare the differences to
        # we will take it as the biggest difference in the last 33% of the difference vector divided by some factor
        eps = np.abs(conv_diff_y[-conv_diff_y.size // 3 :]).max() * eps_size
        # detect regions:
        regions = np.where(np.abs(conv_diff_y) < eps)[0] + low_step
        # split regions into regions of points 'split_tolerance' indices apart
        splits = np.split(regions, np.where(np.diff(regions) >= split_tolerance)[0])
        return regions, splits

    def get_ACF_integral(self, ACFKFFilepath, flat_selection="largest"):
        """
        Method used for reading in a single ACF integral that has been calculated

        arguments:
        ACFKFFilepath  - path to the KFFile with the ACF calculation
        flat_selection - the method will attempt to detect flat areas in the integral.
                         This argument controls which of the flats is chosen.
                         Can be one of ['largest', 'all', 'first', 'last'].

        """

        # read in integrals
        reader = KFReader(ACFKFFilepath)
        integrals = np.asarray(reader.read("Integral(1)", "y-axis"))

        # Get the flats and split flats first, these are regions in the ACF integral
        # that are roughly flat and signify converged regions of the ACF. The split_flats are subarrays
        # of flats whose indices are separated by more than split_tolerance
        # we skip the first 1/3 of the steps
        flats, split_flats = self.get_flat_areas(
            integrals, low_step=integrals.size // 3, kernel_size=800, split_tolerance=10
        )
        if flat_selection == "largest":
            largest_flat = split_flats[np.argmax([flat.size for flat in split_flats])]
            average = np.mean(integrals[largest_flat])
            sd = np.std(integrals[largest_flat])
        elif flat_selection == "all":
            average = np.mean(integrals[flats])
            sd = np.std(integrals[flats])
        elif flat_selection == "first":
            average = np.mean(integrals[split_flats[0]])
            sd = np.std(integrals[split_flats[0]])
        elif flat_selection == "last":
            average = np.mean(integrals[split_flats[-1]])
            sd = np.std(integrals[split_flats[-1]])

        # integral is in units au^2 fs, which we want in Pa^2 s
        average = average * Units.conversion_ratio("au", "Pa") ** 2 * 1e-15
        sd = sd * Units.conversion_ratio("au", "Pa") ** 2 * 1e-15
        return average, sd

    def get_viscosities(self, molfiles, temperatures, name, calculation_path=None):
        """
        Main function used to calculate the viscosity at certain temperatures
        This function does everything from scratch, so first an MD run

        directory structure:
        {calculation_path}
        |
        --- {name}
            |
            --- {name}_T{temp1} #these will be the MD run folders
            --- {name}_T{temp2}
            --- etc.
            --- {name}_T{temp1}_analysis #these will be the ACF calculation folders
            --- {name}_T{temp2}_analysis
            --- etc.

        arguments:
        molfiles         - iterable containing file paths to the input molecule systems
        temperatures     - iterable containing the temperatures that will be used for the MD runs
        name             - name of the viscosity calculations (see directory structure)
        calculation_path - folder in which to initialize PLAMS (see directory structure)
        """

        init(path=calculation_path, folder=name)

        # here we make sure PLAMS does parallelization
        Nruns = len(temperatures)
        config.default_jobrunner = JobRunner(parallel=True, maxjobs=multiprocessing.cpu_count())
        config.job.runscript.nproc = 1

        analysisname = name + "_analysis"
        mols = [Molecule(molfile) for molfile in molfiles]  # load in all molecules

        print(f"Beginning viscosity calculations {name}")
        print(f"I will perform {len(temperatures)} MD runs")

        ##### MD RUNS
        print("Starting MD runs ...")
        # start the jobs
        MDresults = [self._get_md_job(mol, T, name + f"_T{T}").run() for T, mol in zip(temperatures, mols)]

        # wait for the jobs to finish and get the rkf files
        MDrkf = [r["ams.rkf"] for r in MDresults if r.ok()]

        #### ANALYSIS
        results = self.analyse_md(MDrkf, do_init=False)
        finish()
        return results

    def plot_viscosities(self, volumes, temps, ACFKFFilepaths, out_path=None, show=False):
        """
        arguments:
        volumes        - iterable containing volumes corresponding to each temperature
        temps          - iterable of MD temperatures
        ACFKFFilepaths - filepath to ACF rkf files
        out_path       - if not None output the plot to this path
        show           - determine whether to show plot as interactive pyplot gui
        """

        # read in the ACF integrals
        reader = [KFReader(KFFilepath) for KFFilepath in ACFKFFilepaths]
        names = [os.path.basename(KFFilepath).split(".")[0] for KFFilepath in ACFKFFilepaths]
        integralsy = [
            np.asarray(r.read("Integral(1)", "y-axis")) * Units.conversion_ratio("au", "Pa") ** 2 * 1e-15
            for r in reader
        ]
        integralsx = [np.asarray(r.read("Integral(1)", "x(1)-axis")) for r in reader]

        kB = Units.constants["k_B"]
        viscosities = [V * 1e-30 / (kB * T[0]) * I for V, T, I in zip(volumes, temps, integralsy)]
        plt.title("Viscosities vs time")
        plt.xlabel(r"$t\quad (fs)$")
        plt.ylabel(r"$\eta\quad (mPa s)$")
        [plt.plot(x, y * 1000, label=n) for x, y, n in zip(integralsx, viscosities, names)]
        plt.legend()
        if not out_path is None:
            plt.savefig(out_path)
        if show:
            plt.show()


if __name__ == "__main__":
    calculator = ViscosityCalculator()

    ##### If you want to run a new calculation
    ##in this case the systems were obtained by doing equilibration simulations to obtain correct densities at the desired temperatures
    name = "Benzene"
    molfiles = [
        "./benzene_bulk_16_250.xyz",
        "./benzene_bulk_16_300.xyz",
        "./benzene_bulk_16_350.xyz",
    ]  # system to do the MD on
    temperatures = [250, 300, 350]

    ##### To run new simulations:
    r = calculator(molfiles, temperatures, name)

    # If you want to reuse older md simulations
    # these files are simply a list of rkf files corresponding to MD simulations
    # the calculator implements a usefull method that will pull MD rkf files from a previous
    # calculation. For file structure see documentation above

    # directory = './plams_workdir/'
    # files = calculator.get_md_KFFiles(directory)
    # r = calculator.analyse_md(files, name='Benzene_analysis', calculation_path=directory)

    ##final printing of the results
    print("All calculations are now done")
    print("Temperature +- SD (K) | Viscosity (mPa s)")
    for t, tsd, v in zip(r["temperatures"], r["temperature_stdevs"], r["viscosities"]):
        print(f"{t: >11.2f} +- {tsd: <6.2f} | {v*1000:.5f}")
